from __future__ import division
import numpy as np


# pythran export load_a3m(str, float)
# pythran export load_a3m(str)
def load_a3m(fasta, max_gap_fraction=0.9):
    """ load alignment with the alphabet used in GaussDCA """
    mapping = {'-': 21, 'A': 1, 'B': 21, 'C': 2, 'D': 3, 'E': 4, 'F': 5,
               'G': 6, 'H': 7, 'I': 8, 'K': 9, 'L': 10, 'M': 11,
               'N': 12, 'O': 21, 'P': 13, 'Q': 14, 'R': 15, 'S': 16, 'T': 17,
               'V': 18, 'W': 19, 'Y': 20,
               'U': 21, 'Z': 21, 'X': 21, 'J': 21}

    # We want to exclude the lowercase, not ignore the uppercase because of gaps.
    lowercase = set('abcdefghijklmnopqrstuvwxyz')

    # Figure out the length of the sequence
    f = open(fasta)
    for line in f:
        if line.startswith('>'):
            continue
        seq_length = len(line.strip())
        break
    else:
        raise RuntimeError('I cannot find the first sequence')
    f.seek(0)

    parsed = []
    for line in f:
        if line.startswith('>'):
            continue
        line = line.strip()
        gap_fraction = line.count('-') / seq_length
        if gap_fraction <= max_gap_fraction:
            parsed.append([mapping.get(ch, 22) for ch in line
                           if ch not in lowercase])

    return np.array(parsed, dtype=np.int8).T



import numpy as np
from scipy import linalg

from . import _gdca
from . import _load_data


def _compute_FN(mJ, n_cols: int, alphabet_size: int):
    FN = np.zeros((n_cols, n_cols), dtype=np.float64)

    s = alphabet_size - 1

    fs = s
    fs2 = s * s

    for i in range(n_cols - 1):
        _row = i * s
        for j in range(i + 1, n_cols):
            _col = j * s

            patch = mJ[_row: _row + s, _col: _col + s]
            total = patch.sum() / fs2
            rows = patch.sum(axis=1) / fs
            columns = patch.sum(axis=0) / fs

            fn_pre = patch - rows[:, None] - columns[None, :] + total
            fn = (fn_pre * fn_pre).sum()

            FN[i, j] = fn
            FN[j, i] = fn

    FN = np.sqrt(FN)
    return FN, _gdca.apc_correction(FN)


def _compute_gdca_scores(alignment, alignment_T, verbose):
    alphabet_size = alignment.max()

    n_cols = alignment_T.shape[1]
    depth = alignment_T.shape[0]

    if verbose:
        print('Prepare inputs')
    covar, meff = _gdca.prepare_covariance(alignment, alignment_T)

    if verbose:
        print('Invert matrix')
    cho = linalg.cho_factor(covar, check_finite=False)
    mJ = linalg.cho_solve(cho, np.eye(covar.shape[0]), check_finite=False, overwrite_b=True)

    if verbose:
        print('Compute Frobenius Norm')
    FN, FN_corr = _compute_FN(mJ, n_cols, alphabet_size)
    results = dict(gdca=FN, gdca_corr=FN_corr, eff_seq=meff, seq=depth)
    return results


def run(path, verbose=False):
    if verbose:
        print('Loading data')
    ali = _load_data.load_a3m(path)

    return _compute_gdca_scores(np.ascontiguousarray(ali), np.ascontiguousarray(ali.T), verbose)


def compute_weights(path, theta=None):
    ali = _load_data.load_a3m(path)
    if theta is None:
        theta = -1.

    return _gdca.compute_weights(np.ascontiguousarray(ali), np.ascontiguousarray(ali.T), theta)



#!/usr/bin/env python
# encoding: UTF-8

from setuptools import setup, find_packages, dist

# Declare the dependency
dist.Distribution(dict(setup_requires='pythran'))

# Pythran modules
from pythran import PythranExtension

module_parse = PythranExtension('gaussdca._load_data', sources=['gaussdca/_load_data.py'])
module_gdca = PythranExtension('gaussdca._gdca', sources=['gaussdca/_gdca.py'],
                               extra_compile_args=['-fopenmp', '-ftree-vectorize'],
                               extra_link_args=['-fopenmp'])

# Main setup:
setup(name='pyGaussDCA', version='0.1.3',
      description='Fast implementation of GaussDCA',
      url='https://github.com/ElofssonLab/pyGaussDCA/',
      author='David Menéndez Hurtado',
      author_email='davidmenhur@gmail.com',
      license='GPLv3',
      packages=find_packages(),
      package_data={'gaussdca.data': ['gaussdca/test/data/*.a3m']},
      include_package_data=True,
      ext_modules=[module_parse, module_gdca],
      install_requires=['numpy', 'scipy', 'pythran>=0.8.6'],
      classifiers=['Programming Language :: Python',
                   'Topic :: Scientific/Engineering :: Bio-Informatics',
                   'Intended Audience :: Science/Research'],
      zip_safe=False)



from __future__ import absolute_import

from . import _load_data
from . import _gdca
from .gaussdca import run, compute_weights

if not hasattr(_load_data, '__pythran__') or not hasattr(_gdca, '__pythran__'):
    import warnings

    warnings.warn(RuntimeWarning('Modules were not compiled, pyGaussDCA will be slow.'))



import numpy as np


# pythran export apc_correction(float[:, :])
def apc_correction(matrix):
    corrected = matrix - np.mean(matrix, axis=0)[None, :] * np.mean(matrix, axis=1)[:, None] / np.mean(matrix)
    np.fill_diagonal(corrected, 0.0)
    return corrected


def _compute_theta(alignment):
    alignment_depth = alignment.shape[1]
    n_cols = alignment.shape[0]

    meanfracid = 0.0
    for i in range(n_cols):
        match_groups = np.bincount(alignment[i, :])
        # N choose 2 has a nice polynomial expansion:
        meanfracid += np.sum(match_groups * (match_groups - 1) / 2)
    meanfracid /= (0.5 * alignment_depth * (alignment_depth - 1) * n_cols)
    theta = min(0.5, 0.38 * 0.32 / meanfracid)
    return theta


def _compute_weights(alignment, theta: float, n_cols: int, depth: int):
    _thresh = np.floor(theta * n_cols)
    counts = np.ones(depth, dtype=np.float64)

    if theta == 0:
        Meff = depth
        return Meff, counts

    # omp parallel for schedule(dynamic, 10)
    for i in range(depth - 1):
        this_vec = alignment[i, :]
        for j in range(i + 1, depth):
            _dist = np.sum(this_vec != alignment[j, :])

            if _dist < _thresh:
                counts[i] += 1.
                counts[j] += 1.

    weights = 1. / counts
    Meff = weights.sum()
    return Meff, weights


def _compute_freqs(alignment, n_cols: int, depth: int, q: int, W):
    s = q - 1
    expanded_cols = n_cols * s

    Pi = np.zeros(expanded_cols, dtype=np.float64)
    Pij = np.zeros((expanded_cols, expanded_cols), dtype=np.float64)

    for i in range(n_cols):
        i0 = i * s
        for k in range(depth):
            a = alignment[i, k]
            if a != q:
                Pi[i0 + a - 1] += W[k]

    for i in range(n_cols):
        i0 = i * s
        j0 = i0
        for j in range(i, n_cols):
            for k in range(depth):
                a = alignment[i, k]
                b = alignment[j, k]

                if a != q and b != q:
                    Pij[i0 + a - 1, j0 + b - 1] += W[k]
            j0 = j0 + s

    Meff = W.sum()
    Pi /= Meff
    Pij = np.maximum(Pij, Pij.T)
    Pij /= Meff

    return Pi, Pij


def _add_pseudocount(Pi_true, Pij_true, pc: float, n_cols: int, q: int):
    pcq = pc / q
    s = q - 1

    Pi = (1 - pc) * Pi_true + pcq
    Pij = (1 - pc) * Pij_true + pcq / q

    i0 = 0
    for i in range(n_cols):
        # Per block correction
        Pij[i0: i0 + s, i0:i0 + s] = (1 - pc) * Pij_true[i0: i0 + s, i0:i0 + s]

        for alpha in range(s):
            x = i0 + alpha
            Pij[x, x] += pcq
        i0 += s

    return Pi, Pij


def _compute_covar(alignment, weights, pseudocount):
    depth = alignment.shape[1]
    n_cols = alignment.shape[0]

    alphabet_size = alignment.max()
    # calculate theta, sequence weights, and frequencies
    Pi_true, Pij_true = _compute_freqs(alignment, n_cols, depth, alphabet_size, weights)

    # adjust frequencies with pseudocounts
    Pi, Pij = _add_pseudocount(Pi_true, Pij_true, pseudocount, n_cols, alphabet_size)

    # generate covariance matrix
    Pi_np = Pi[:, np.newaxis]
    covar = Pij - (Pi_np * Pi_np.T)

    return covar


# pythran export prepare_covariance(int8[:, :],int8[:, :], float)
# pythran export prepare_covariance(int8[:, :], int8[:, :])
def prepare_covariance(alignment, alignment_T, pseudocount=0.8):
    n_cols = alignment_T.shape[1]
    depth = alignment_T.shape[0]

    theta = _compute_theta(alignment)
    meff, weights = _compute_weights(alignment_T, theta, n_cols, depth)
    covar = _compute_covar(alignment, weights, pseudocount)
    return covar, meff


# pythran export compute_weights(int8[:, :], int8[:, :], float)
def compute_weights(alignment, alignment_T, theta):
    if theta <= 0.0:
        theta = _compute_theta(alignment)
    n_cols = alignment_T.shape[1]
    depth = alignment_T.shape[0]
    meff, weights = _compute_weights(alignment_T, theta, n_cols, depth)
    return weights


# pythran export compute_FN(float64[:, :], int, int8)
# pythran export compute_FN(float[:, :], int, int8)
# pythran export compute_FN(float64[::, :], int, int8)
# pythran export compute_FN(float[::, :], int, int8)
def compute_FN(mJ, n_cols: int, alphabet_size: int):
    FN = np.zeros((n_cols, n_cols), dtype=np.float64)

    s = alphabet_size - 1

    fs = s
    fs2 = s * s

    for i in range(n_cols - 1):
        _row = i * s
        for j in range(i + 1, n_cols):
            _col = j * s

            patch = mJ[_row: _row + s, _col: _col + s]
            total = patch.sum() / fs2
            rows = patch.sum(axis=1) / fs
            columns = patch.sum(axis=0) / fs

            fn_pre = patch - rows[:, None] - columns[None, :] + total
            fn = (fn_pre * fn_pre).sum()

            FN[i, j] = fn
            FN[j, i] = fn

    FN = np.sqrt(FN)
    return FN, apc_correction(FN)
